unit mbox;

(*

  Este es un primer intento de encapsular mboxes estndar de Unix en
  componentes. El parser funciona en mis archivos de correo y grupos de
  discusin en Linux de Netscape y Mutt, as como en mis archivos de correo
  en Netscape en Windows. No hay ninguna garanta de que servir con sus
  archivos, pero no pierde usted nada con tratar.

  Aunque es de lectura unicamente y no estoy escribiendo nada al archio, no le
  garantizo la integridad de sus datos - haga copias.

  Nota: El TMailMessage que utilizo aqu es mi propio TMailMessage, y NO el
  del componente POP3. Cuando comenc a escribir esto no me haba fijado que
  el componente tambin tena una estructura llamada TMailMessage. As que
  cuando lo utilice en otras unidades, voy a calificar la estructura con el
  nombre de la unidad, por ejemplo "mbox.TMailMessage".

*)

interface
  uses classes, SysUtils;

type

  TUpdateProc = procedure( Current, Max : Integer; CustomMsg : String ) of object;

  TMailMessage = class
  private
    FMailFrom    : String;
    FMailTo      : String;
    FMailSubject : String;
    FMailDate    : String;
    FMailBody    : TStrings;
    FMailMsgID   : String;
    FAttributes  : TStrings;
  public
    property MailFrom : String read FMailFrom write FMailFrom;
    property MailTo   : String read FMailTo write FMailTo;
    property MailSubject : String read FMailSubject write FMailSubject;
    property MailDate    : String read FMailDate    write FMailDate;
    property MailBody : TStrings read FMailBody write FMailBody;
    property MailMsgId    : String read FMailMsgID    write FMailMsgId;
    property Attributes   : TStrings read FAttributes write FAttributes;
    constructor Create;
    destructor Destroy; override;
  end;

  TMessageList = class(TList)
  protected
    function GetItem( ItemNo : Integer ) : TMailMessage;
    procedure SetItem( ItemNo : Integer; Item : TMailMessage );
  public
    property Items[Index: Integer] : TMailMessage read GetItem write SetItem;
    procedure AddMessage( Item : TMailMessage );
    procedure DeleteMessage( Index : Integer );
  end;

  TMBox = class
  private
    FHandle       : TextFile;
  protected
    FMBoxName    : String;       // Nombre de archivo del Mbox
    FUpdateProc  : TUpdateProc;  // Procedimiento que usaremos para actualizar la interfaz de usuario
    FMessages    : TMessageList; // La lista de Mensajes
    FValidHdrs   : TStrings;     // Lista de Encabezados Vlidos
    FDataFile    : TStrings;     // El Archivo de datos.
    procedure ReadDataFile;      // Lee el archivo de datos
    function getMBoxName : String;
    procedure setMBoxName(Value : String);  // Get/Set para el nombre de archivo del MBox
    function getUpdateProc : TUpdateProc;
    procedure setUpdateProc(Value : TUpdateProc);  // Get/Set para el procedimiento de actualizacin
    function ReadHeaders( Data : TStrings; i : Integer; TempMsg : TMailMessage ): Integer; // Lee los encabezados en un TMailMessage
    function ReadBody( Data : TStrings; i : Integer; TempMsg : TMailMessage ): Integer;    // Lee el cuerpo
    function IsInHeader( S : String): Boolean;
    function InValidHeaders(S : String): Boolean;
    function StartsWithTaborSpaces( S : String): Boolean;
  public
    procedure FillMessages;
    constructor Create;
    destructor Destroy; override;
    property MessageList : TMessageList read FMessages write FMessages;
  published
    property MBoxName : String read getMBoxName write setMBoxName;
    property UpdateProc : TUpdateProc read GetUpdateProc write SetUpdateProc;
  end;

  function MboxFileSize(MBoxName : String) : Integer;
  // Pon en un Strings preinicializado todos los nombres de archivos tipo MBox
  procedure GetMBoxNames( Directory : String; Strings : TStrings );
  // verifica si un nombre de archivo es un mbox, revisando los primeros bytes
  function IsanMBox( FileName : String ) : Boolean;
  // Mi ReadLine que funciona en archivos Unix y DOS. 
  function RealReadln( var F : TextFile ): String;

const
  // Lista de los identificadores de atributos validos para un mbox
  MBOX_FROMPFX = 'FROM ';
  MBOX_HDR_RETURNPATH = 'RETURN-PATH';
  MBOX_HDR_RECEIVED   = 'RECEIVED';
  MBOX_HDR_INREPLY    = 'IN-REPLY-TO';
  MBOX_HDR_SENDER     = 'SENDER';
  MBOX_HDR_FROM       = 'FROM';
  MBOX_MSG_ID         = 'MESSAGE-ID';
  MBOX_HDR_DATE       = 'DATE';
  MBOX_HDR_MIMEVERSION   = 'MIME-VERSION';
  MBOX_HDR_TO            = 'TO';
  MBOX_HDR_CC            = 'CC';
  MBOX_HDR_FCC            = 'FCC';
  MBOX_HDR_BCC            = 'BCC';
  MBOX_HDR_ORGANIZATION  = 'ORGANIZATION';
  MBOX_HDR_SUBJECT       = 'SUBJECT';
  MBOX_HDR_NEWSGROUPS    = 'NEWSGROUPS';
  MBOX_HDR_REFERENCES    = 'REFERENCES';
  MBOX_HDR_CONTENT_TYPE  = 'CONTENT-TYPE';
  MBOX_HDR_CONTENT_BASE  = 'CONTENT-BASE';
  MBOX_HDR_CONTENT_DISPOSITION = 'CONTENT-DISPOSITION';
  MBOX_HDR_CONTENT_XFER_ENC = 'CONTENT-TRANSFER-ENCODING';
  MBOX_HDR_REPLYTO       = 'REPLY-TO';
  MBOX_HDR_ERRORSTO      = 'ERRORS-TO';
  // Cabeceras extendidas que deseo reconocer.
  MBOX_HDR_X_MAILER    = 'X-MAILER: ';
  MBOX_HDR_X_ACCEPT_LANG = 'X-ACCEPT-LANGUAGE: ';
  XT_ATTR_PREFIX         = 'X-';  // Prefijo para una cabecera extendida
  A_SEP                  = ': ';  // Separador de Atributos/Valores.

implementation

// MessageList hereda de TList y solo permite TMailMessages.
procedure TMessageList.AddMessage( Item : TMailMessage );
begin
  Add( Item );
end;

procedure TMessageList.DeleteMessage( Index : Integer );
begin
  Delete( Index );
end;

function TMessageList.GetItem( ItemNo : Integer ) : TMailMessage;
begin
  Result := TMailMessage( inherited Items[ItemNo] );
end;

procedure TMessageList.SetItem( ItemNo : Integer; Item : TMailMessage );
begin
  Items[ItemNo] := Item;
end;

// Implementacin del Mensaje.

constructor TMailMessage.Create;
begin

  FAttributes := TStringList.Create;
  FMailBody   := TStringList.Create;

end;

destructor TMailMessage.Destroy;
begin

  FAttributes.Free;
  FMailBody.Free;
  inherited;

end;



// Implementacin de una caja de correo.

constructor TMBox.Create;
begin

  FUpdateProc := Nil;
  FMessages := TMessageList.Create;
  FValidHdrs := TStringList.Create;
  FDataFile  := TStringList.Create;

  // Lista de Cabeceras vlidas en un mensaje
  FValidHdrs.Add(MBOX_FROMPFX);
  FValidHdrs.Add(MBOX_HDR_RETURNPATH);
  FValidHdrs.Add(MBOX_HDR_INREPLY);
  FValidHdrs.Add(MBOX_HDR_RECEIVED);
  FValidHdrs.Add(MBOX_HDR_FROM);
  FValidHdrs.Add(MBOX_MSG_ID);
  FValidHdrs.Add(MBOX_HDR_SENDER);
  FValidHdrs.Add(MBOX_HDR_DATE);
  FValidHdrs.Add(MBOX_HDR_X_MAILER);
  FValidHdrs.Add(MBOX_HDR_X_ACCEPT_LANG);
  FValidHdrs.Add(MBOX_HDR_MIMEVERSION);
  FValidHdrs.Add(MBOX_HDR_TO);
  FValidHdrs.Add(MBOX_HDR_SUBJECT);
  FValidHdrs.Add(MBOX_HDR_REFERENCES);
  FValidHdrs.Add(MBOX_HDR_NEWSGROUPS);
  FValidHdrs.Add(MBOX_HDR_CONTENT_TYPE);
  FValidHdrs.Add(MBOX_HDR_CONTENT_BASE);
  FValidHdrs.Add(MBOX_HDR_CONTENT_XFER_ENC);
  FValidHdrs.Add(MBOX_HDR_REPLYTO);
  FValidHdrs.Add(MBOX_HDR_ERRORSTO);
  FValidHdrs.Add(MBOX_HDR_CONTENT_DISPOSITION);
  FValidHdrs.Add(MBOX_HDR_ORGANIZATION);
  FValidHdrs.Add(MBOX_HDR_CC);
  FValidHdrs.Add(MBOX_HDR_FCC);
  FValidHdrs.Add(MBOX_HDR_BCC);

  FValidHdrs.Add(A_SEP);

end;

destructor TMBox.Destroy;
var
  i : Integer;
  msg : TMailMessage;
begin

  for i := FMessages.Count -1 downto 0 do begin
    msg := FMessages.Items[i];
    FMessages.DeleteMessage(i);
    msg.Free;
  end;
  FMessages.Free;
  FValidHdrs.Free;
  FUpdateProc := Nil;
  inherited;

end;

procedure TMBox.setMBoxName(Value : String);
begin

  // mBoxName is a file name. Here we should
  // try to find out if the file exists and is a mailbox.
  if not FileExists(Value) then
     raise Exception.Create('Archivo no existente!');
  if not IsAnMBox(Value) then
     raise Exception.Create('Este archivo no es un archivo de correo (al menos no en formato "mbox").');

  FMBoxName := Value;

end;

function TMBox.getMBoxName: String;
begin
  Result := FMBoxName;
end;

function TMBox.getUpdateProc : TUpdateProc;
begin

   Result := FUpdateProc;

end;

procedure TMBox.setUpdateProc(Value : TUpdateProc);
begin

   FUpdateProc := Value;

end;

procedure TMBox.ReadDataFile;
var
  F : TextFile;
  S : String;
  iCurrent, iMax : Integer;
begin

  iCurrent := 0;
  iMax := MboxFileSize(MBoxName);
  if @FUpdateProc <> Nil then FUpdateProc(iCurrent, iMax, 'Abriendo archivo de correo para '+MBoxName+'...');
  AssignFile(F, MBoxName);
  FDataFile.Clear;
  Reset(F);
  while not Eof(F) do begin  // Lee todo el archivo
    if @FUpdateProc <> Nil then FUpdateProc(iCurrent, iMax, 'Leyendo archivo de datos...');
    S := RealReadLn(F);
    Inc(iCurrent, Length(S)+1);
    FDataFile.Add(S);
  end;

end;

function TMBox.InValidHeaders(S : String): Boolean;
var
  i : Integer;
begin

  Result := False;
  i := 0;
  while Result = False do begin
    Result := UpperCase(Copy(s, 1, Length(FValidHdrs[i]))) = FValidHdrs[i];
    Inc(i);
    if i = FValidHdrs.Count then exit;
  end;


end;

function TMBox.IsInHeader( S : String): Boolean;
begin

  // Esto nos dir si una lnea en especfico es parte de la cabecera dependiendo del criterio:
  //  1) La lnea comienza con un atributo extendido
  //  2) La lnea Comienza con un atributo verdadero
  // Slo es exacta dependiendo del contexto. Vea "ReadHeaders" para mayor informacin
  Result := False;
  Result := Pos(XT_ATTR_PREFIX, S) = 1;
  if Result = False then Result := InValidHeaders(S);

end;

function TMBox.StartsWithTaborSpaces( S : String): Boolean;
begin

   Result := ( Pos(' ', S) = 1 ) or ( Pos(#9, S) = 1) ;

end;


// Esta funcin lee los encabezados y regresa con el numero de lnea (a partir de "i")
//  para la cual ya no hay ms encabezados, para que la funcin que la llama sepa
// desde dnde continuar leyendo el archivo. Un valor de -1 querr decir que no
// hubo encabezados.
function TMbox.ReadHeaders( Data : TStrings; i : Integer; TempMsg : TMailMessage ): Integer;
var
  iPos : Integer;
  bContinue : Boolean;
  bMultiLineSection : Boolean;
  sCurrent : String;
begin

  bContinue := True;
  iPos := 0;
  bMultiLineSection := False;
  while bContinue do begin
    sCurrent := Data[i+iPos];
    if IsInHeader(sCurrent) then begin
      bMultiLineSection :=   ( UpperCase(Copy(sCurrent, 1, Length(MBOX_HDR_RECEIVED))) = MBOX_HDR_RECEIVED)
                           or ( UpperCase(Copy(sCurrent, 1, Length(MBOX_HDR_CONTENT_TYPE))) = MBOX_HDR_CONTENT_TYPE)
                           or ( Copy(sCurrent, Length(sCurrent), 1) = ',' )
                           or ( Copy(sCurrent, Length(sCurrent), 1) = ';' )
                           or ( Copy(sCurrent, Length(sCurrent), 1) = '\' )
                           or ( Copy(sCurrent, Length(sCurrent), 1) = '=' )
                           ;
      if Pos(A_SEP, sCurrent)>0 then sCurrent[Pos(A_SEP, sCurrent)] := '=';
      TempMsg.Attributes.Add(sCurrent);
      Inc(iPos);
    end
    else
    begin
       if bMultiLineSection  then begin
          if StartsWithTaborSpaces(sCurrent)
             then begin
               if TempMsg.Attributes.Count > 0 then
                 TempMsg.Attributes[TempMsg.Attributes.Count-1] := TempMsg.Attributes[TempMsg.Attributes.Count-1]+#13+#10+sCurrent;
               Inc(iPos);
             end else bContinue := False;
       end else bContinue := False;
    end
  end;
  Result := iPos;
  // Al final, checa en el TMailMessage a ver si hay un atributo obligatorio. Si no existe, devuelve -1.
  if TempMsg.Attributes.Values[MBOX_HDR_SUBJECT] = '' then begin
    TempMsg.Attributes.Clear;
    Result := -1;
  end;

end;


// Esta funcin checa el cuerpo del mensaje hasta que el prximo
// encabezado comienza. Los parmetros y resultado son iguales al
// ReadHeaders. Si usted decide manejar archivos aadidos, aqu es
// donde podra tratar de averiguar acerca de ellos.
function TMBox.ReadBody( Data : TStrings; i : Integer; TempMsg : TMailMessage ): Integer;
var
  bContinue : Boolean;
  sCurrent : String;
  sHdrCheck : String;
  iPos : Integer;
begin

  iPos := 0;
  bContinue := True;
  while bContinue do begin
    if i+iPos > Data.Count-1 then break;
    sCurrent := Data[i+iPos];
    if UpperCase(Copy(sCurrent, 1, Length(MBOX_FROMPFX))) = MBOX_FROMPFX then begin
       // Wait until the next 'from ' and then if found check 2 lines ahead to see if it's a header
       sHdrCheck := Data[i+iPos+1];
       bContinue := not IsInHeader(sHdrCheck);
       sHdrCheck := Data[i+iPos+2];
       bContinue := bContinue or ( not IsInHeader(sHdrCheck) );
       if not bContinue then begin
         Dec(iPos);
         break;
       end;
    end;
    TempMsg.MailBody.Add(sCurrent);
    Inc(iPos);
  end;
  Result := iPos;

end;


procedure TMBox.FillMessages;
var
  TempMSG : TMailMessage;  
  i, h, iCurrent, iMax, iHowDeep : Integer;
  sCurrent : String;
  AttrState : Integer;  // 1 = Encabezados, 2 = Cuerpo.
begin

  if FDataFile.Count = 0 then ReadDataFile;

  // Cada Lnea DEBE ser parte de un TMailMessage, ya sea en los atributos o
  // en el cuerpo. As que comenzamos por crear un TMailMessage y aadir
  // atributos. Cuando terminamos de aadir atributos, aadimos el cuerpo. Si
  // el lector del cuerpo comienza a encontrar atributos de nuevo regresar
  // dicindomnos cuntas lneas ley, lo cual quiere decir que la lnea siguiente
  // es un atributo de nuevo.
  AttrState := 1;  // Comienza con estatus de "encabezado"
  iCurrent := 0;
  i:= 0;
  TempMsg := Nil;

  iMax := FDataFile.Count;
  while i < iMax do begin
    iCurrent := i;
    if @FUpdateProc <> Nil then FUpdateProc(iCurrent, iMax, 'Parsing Data File...');
    sCurrent := FDataFile[i];
    // Si nos toca leer encabezados
    If AttrState = 1 then begin
      if TempMsg <> Nil then MessageList.AddMessage(TempMsg);
      TempMsg := TMailMessage.Create;
      h := ReadHeaders( FDataFile, i, TempMsg );
      if h = - 1 then begin
        if MessageList.Count > 0 then MessageList.Delete(MessageList.Count-1);
        TempMsg.Free;
        TempMsg := Nil;
        AttrState := 2;
      end else begin
        TempMsg.MailFrom := TempMSG.Attributes.Values[MBOX_HDR_FROM];
        TempMsg.MailSubject := TempMSG.Attributes.Values[MBOX_HDR_SUBJECT];
        TempMsg.MailDate := TempMSG.Attributes.Values[MBOX_HDR_DATE];
        TempMsg.MailMsgId := TempMSG.Attributes.Values[MBOX_MSG_ID];
        i := i + h;
        sCurrent := FDataFile[i];
        AttrState := 2;
      end;
    end;
    // Si nos toca leer el cuerpo
    if AttrState = 2 then begin
       h := ReadBody( FDataFile, i, TempMsg );
       if h <> -1 then begin
         AttrState := 1;
         i := i + h;
       end else begin
         raise Exception.Create('Cuerpo vaco (ni siquiera una lnea vaca)?? No puedo reconocer este archivo!');
         //TempMsg.Free;
         // AttrState := 1;
         // TempMsg := Nil;
         
       end;
    end;
    Inc(i);
  end;
  // Al final, si tenemos un mensaje que se ley, aadelo a la lista
  // porque es el ltimo mensaje que lemos.
  if TempMsg <> Nil then MessageList.AddMessage(TempMsg);
  
end;

function MboxFileSize(MBoxName : String) : Integer;
var
  rec : TSearchRec;
begin

  Result := 0;
  if FindFirst(MBoxName, faAnyFile, rec) = 0 then Result := Rec.Size;

end;


(*
  Esta rutina pone los nombres de archivo de todos los mboxes
  (dentro de un directorio) en una TStrings (que debi haber
  sido inicializada con anterioridad). Verifica que el MBox
  est en formato MBOX (va IsanMBox), para que no haya problema
  si le damos un directorio con algunos mboxes y otros archivos
  miscelneos.
*)
procedure GetMBoxNames( Directory : String; Strings : TStrings );
var
  rec : TSearchRec;
  i: Integer;
begin

  // Borra el Strings.
  Strings.Clear;
  i := FindFirst(Directory+'/*', faAnyFile, rec);
  if i = 0 then begin
    while i = 0 do begin
      // Por cada archivo que no sea un directorio
     if ( rec.Attr and faDirectory ) = 0 then
     if IsanMBox(Directory+'/'+rec.Name ) then Strings.Add(rec.Name);
     i := FindNext(rec)

    end;
    FindClose(rec);
  end;


end;


// Averigua si un archivo es un MBox
function IsanMBox( FileName : String ) : Boolean;
var
  F: TextFile;
  ptr : PChar;
  iFromLength : Integer;
  i : Integer;
  sDeleteme : String;
begin


  Result := False;
  iFromLength := Length(MBOX_FROMPFX);
  ptr := AllocMem(iFromLength+1);  // +1 porque el ltimo es #0, que es el delimitador de fin de cadena en PChars.
  try

    ptr[iFromLength] := #0;
    begin
      try
        AssignFile( F, FileName );
        Reset(F);
        // Lea una cadena de longitod iFromLength, byte por byte
        for i := 0 to iFromLength-1 do
           Read(F, ptr[i]);
        if UpperCase(ptr) = MBOX_FROMPFX then Result := True;
        sDeleteme := RealReadLn(F);
      except
        Close(f);
      end;
//         Close(f);
    end;
  finally

    FreeMem(Ptr);
  end;


end;

(*
  RealReadLn es un ReadLn independiente de la plataforma. Puede
  leer archivos con CR y CR/LF sin quejarse ni leer de ms.
*)
function RealReadln( var F : TextFile ): String;
var
  ptr : PChar;
  iPos : Integer;
begin

  Result := '';
  ptr := AllocMem(2);  // +1 because the last one is #0
  try

    ptr[1] := #0;
    Read(F, ptr[0]);
    while ( ptr <> #$A ) AND ( ptr <> #$1A ) and ( ptr<>#$0D ) do begin
      Result := Result + ptr;
      Read(F, ptr[0]);
    end;
    // Esto nos permite leer archivos de MS-DOS tambin, porque mueve el puntero
    // un caracter extra (el LF)
    if ptr = #$0D then Read(F, ptr[0]);

  finally
    FreeMem(Ptr);
  end;

end;


end.
